@@ -17,9 +17,6 @@
  */
 package org.apache.orc.impl;
 
-import java.time.chrono.ChronoLocalDate;
-import java.time.chrono.Chronology;
-import java.time.chrono.IsoChronology;
 import org.apache.hadoop.hive.common.type.HiveDecimal;
 import org.apache.hadoop.hive.serde2.io.DateWritable;
 import org.apache.hadoop.hive.serde2.io.HiveDecimalWritable;
@@ -38,11 +35,14 @@ import org.apache.orc.OrcProto;
 import org.apache.orc.StringColumnStatistics;
 import org.apache.orc.TimestampColumnStatistics;
 import org.apache.orc.TypeDescription;
+import org.threeten.extra.chrono.HybridChronology;
 
 import java.sql.Date;
 import java.sql.Timestamp;
+import java.time.chrono.ChronoLocalDate;
+import java.time.chrono.Chronology;
+import java.time.chrono.IsoChronology;
 import java.util.TimeZone;
-import org.threeten.extra.chrono.HybridChronology;
 
 
 public class ColumnStatisticsImpl implements ColumnStatistics {
@@ -64,10 +64,7 @@ public class ColumnStatisticsImpl implements ColumnStatistics {
     if (hasNull != that.hasNull) {
       return false;
     }
-    if (bytesOnDisk != that.bytesOnDisk) {
-      return false;
-    }
-    return true;
+    return bytesOnDisk == that.bytesOnDisk;
   }
 
   @Override
@@ -155,11 +152,7 @@ public class ColumnStatisticsImpl implements ColumnStatistics {
 
       BooleanStatisticsImpl that = (BooleanStatisticsImpl) o;
 
-      if (trueCount != that.trueCount) {
-        return false;
-      }
-
-      return true;
+      return trueCount == that.trueCount;
     }
 
     @Override
@@ -294,11 +287,7 @@ public class ColumnStatisticsImpl implements ColumnStatistics {
       if (maximum != that.maximum) {
         return false;
       }
-      if (sum != that.sum) {
-        return false;
-      }
-
-      return true;
+      return sum == that.sum;
     }
 
     @Override
@@ -503,11 +492,7 @@ public class ColumnStatisticsImpl implements ColumnStatistics {
       if (hasMinimum != that.hasMinimum) {
         return false;
       }
-      if (overflow != that.overflow) {
-        return false;
-      }
-
-      return true;
+      return overflow == that.overflow;
     }
 
     @Override
@@ -661,11 +646,7 @@ public class ColumnStatisticsImpl implements ColumnStatistics {
       if (Double.compare(that.maximum, maximum) != 0) {
         return false;
       }
-      if (Double.compare(that.sum, sum) != 0) {
-        return false;
-      }
-
-      return true;
+      return Double.compare(that.sum, sum) == 0;
     }
 
     @Override
@@ -918,11 +899,7 @@ public class ColumnStatisticsImpl implements ColumnStatistics {
       if (minimum != null ? !minimum.equals(that.minimum) : that.minimum != null) {
         return false;
       }
-      if (maximum != null ? !maximum.equals(that.maximum) : that.maximum != null) {
-        return false;
-      }
-
-      return true;
+      return maximum != null ? maximum.equals(that.maximum) : that.maximum == null;
     }
 
     @Override
@@ -1082,11 +1059,7 @@ public class ColumnStatisticsImpl implements ColumnStatistics {
 
       BinaryStatisticsImpl that = (BinaryStatisticsImpl) o;
 
-      if (sum != that.sum) {
-        return false;
-      }
-
-      return true;
+      return sum == that.sum;
     }
 
     @Override
@@ -1251,11 +1224,7 @@ public class ColumnStatisticsImpl implements ColumnStatistics {
       if (maximum != null ? !maximum.equals(that.maximum) : that.maximum != null) {
         return false;
       }
-      if (sum != null ? !sum.equals(that.sum) : that.sum != null) {
-        return false;
-      }
-
-      return true;
+      return sum != null ? sum.equals(that.sum) : that.sum == null;
     }
 
     @Override
@@ -1631,11 +1600,7 @@ public class ColumnStatisticsImpl implements ColumnStatistics {
       if (minimum != that.minimum) {
         return false;
       }
-      if (maximum != that.maximum) {
-        return false;
-      }
-
-      return true;
+      return maximum == that.maximum;
     }
 
     @Override
