@@ -17,12 +17,7 @@
  */
 package org.apache.orc.impl;
 
-import java.time.LocalDate;
-import java.time.LocalDateTime;
-import java.time.LocalTime;
-import java.time.ZoneOffset;
-import java.time.chrono.ChronoLocalDate;
-import java.time.format.DateTimeFormatter;
+import org.apache.commons.lang3.ArrayUtils;
 import org.apache.hadoop.fs.FSDataInputStream;
 import org.apache.hadoop.fs.Path;
 import org.apache.hadoop.hive.common.type.HiveDecimal;
@@ -44,6 +39,7 @@ import org.apache.orc.DoubleColumnStatistics;
 import org.apache.orc.IntegerColumnStatistics;
 import org.apache.orc.OrcConf;
 import org.apache.orc.OrcFile;
+import org.apache.orc.OrcFilterContext;
 import org.apache.orc.OrcProto;
 import org.apache.orc.Reader;
 import org.apache.orc.RecordReader;
@@ -51,8 +47,12 @@ import org.apache.orc.StringColumnStatistics;
 import org.apache.orc.StripeInformation;
 import org.apache.orc.TimestampColumnStatistics;
 import org.apache.orc.TypeDescription;
+import org.apache.orc.filter.BatchFilter;
+import org.apache.orc.impl.filter.FilterFactory;
 import org.apache.orc.impl.reader.ReaderEncryption;
 import org.apache.orc.impl.reader.StripePlanner;
+import org.apache.orc.impl.reader.tree.BatchReader;
+import org.apache.orc.impl.reader.tree.TypeReader;
 import org.apache.orc.util.BloomFilter;
 import org.apache.orc.util.BloomFilterIO;
 import org.slf4j.Logger;
@@ -61,11 +61,20 @@ import org.slf4j.LoggerFactory;
 import java.io.IOException;
 import java.math.BigDecimal;
 import java.sql.Timestamp;
+import java.time.LocalDate;
+import java.time.LocalDateTime;
+import java.time.LocalTime;
+import java.time.ZoneOffset;
+import java.time.chrono.ChronoLocalDate;
+import java.time.format.DateTimeFormatter;
 import java.util.ArrayList;
 import java.util.Arrays;
 import java.util.Date;
 import java.util.List;
+import java.util.SortedSet;
 import java.util.TimeZone;
+import java.util.TreeSet;
+import java.util.function.Consumer;
 
 public class RecordReaderImpl implements RecordReader {
   static final Logger LOG = LoggerFactory.getLogger(RecordReaderImpl.class);
@@ -80,17 +89,26 @@ public class RecordReaderImpl implements RecordReader {
   private final boolean[] fileIncluded;
   private final long rowIndexStride;
   private long rowInStripe = 0;
+  // position of the follow reader within the stripe
+  private long followRowInStripe = 0;
   private int currentStripe = -1;
   private long rowBaseInStripe = 0;
   private long rowCountInStripe = 0;
-  private final TreeReaderFactory.TreeReader reader;
+  private final BatchReader reader;
   private final OrcIndex indexes;
+  // identifies the columns requiring row indexes
+  private final boolean[] rowIndexColsToRead;
   private final SargApplier sargApp;
   // an array about which row groups aren't skipped
   private boolean[] includedRowGroups = null;
   private final DataReader dataReader;
   private final int maxDiskRangeChunkLimit;
   private final StripePlanner planner;
+  // identifies the type of read, ALL(read everything), LEADERS(read only the filter columns)
+  private final TypeReader.ReadPhase startReadPhase;
+  // identifies that follow columns bytes must be read
+  private boolean needsFollowColumnsRead;
+  private final boolean noSelectedVector;
   // identifies whether the file has bad bloom filters that we should not use.
   private final boolean skipBloomFilters;
   static final String[] BAD_CPP_BLOOM_FILTER_VERSIONS = {
@@ -102,20 +120,55 @@ public class RecordReaderImpl implements RecordReader {
    *
    * @param evolution the mapping from reader to file schema
    * @param columnName  the fully qualified column name to look for
-   * @return the file column number or -1 if the column wasn't found
+   * @return the file column number or -1 if the column wasn't found in the file schema
+   * @throws IllegalArgumentException if the column was not found in the reader schema
    */
   static int findColumns(SchemaEvolution evolution,
                          String columnName) {
+    TypeDescription fileColumn = findColumnType(evolution, columnName);
+    return fileColumn == null ? -1 : fileColumn.getId();
+  }
+
+  static TypeDescription findColumnType(SchemaEvolution evolution, String columnName) {
     try {
       TypeDescription readerColumn = evolution.getReaderBaseSchema().findSubtype(
-          columnName, evolution.isSchemaEvolutionCaseAware);
-      TypeDescription fileColumn = evolution.getFileType(readerColumn);
-      return fileColumn == null ? -1 : fileColumn.getId();
+        columnName, evolution.isSchemaEvolutionCaseAware);
+      return evolution.getFileType(readerColumn);
     } catch (IllegalArgumentException e) {
-      if (LOG.isDebugEnabled()){
-        LOG.debug("{}", e.getMessage());
-      }
-      return -1;
+      throw new IllegalArgumentException("Filter could not find column with name: " +
+                                         columnName + " on " + evolution.getReaderBaseSchema(),
+                                         e);
+    }
+  }
+
+  /**
+   * Given a column name such as 'a.b.c', this method returns the column 'a.b.c' if present in the
+   * file. In case 'a.b.c' is not found in file then it tries to look for 'a.b', then 'a'. If none
+   * are present then it shall return null.
+   *
+   * @param evolution the mapping from reader to file schema
+   * @param columnName the fully qualified column name to look for
+   * @return the file column type or null in case none of the branch columns are present in the file
+   * @throws IllegalArgumentException if the column was not found in the reader schema
+   */
+  static TypeDescription findMostCommonColumn(SchemaEvolution evolution, String columnName) {
+    try {
+      TypeDescription readerColumn = evolution.getReaderBaseSchema().findSubtype(
+        columnName, evolution.isSchemaEvolutionCaseAware);
+      TypeDescription fileColumn;
+      do {
+        fileColumn = evolution.getFileType(readerColumn);
+        if (fileColumn == null) {
+          readerColumn = readerColumn.getParent();
+        } else {
+          return fileColumn;
+        }
+      } while (readerColumn != null);
+      return null;
+    } catch (IllegalArgumentException e) {
+      throw new IllegalArgumentException("Filter could not find column with name: " +
+                                         columnName + " on " + evolution.getReaderBaseSchema(),
+                                         e);
     }
   }
 
@@ -130,10 +183,15 @@ public class RecordReaderImpl implements RecordReader {
                             List<PredicateLeaf> sargLeaves,
                             SchemaEvolution evolution) {
     int[] result = new int[sargLeaves.size()];
-    Arrays.fill(result, -1);
-    for(int i=0; i < result.length; ++i) {
-      String colName = sargLeaves.get(i).getColumnName();
-      result[i] = findColumns(evolution, colName);
+    for (int i = 0; i < sargLeaves.size(); ++i) {
+      int colNum = -1;
+      try {
+        String colName = sargLeaves.get(i).getColumnName();
+        colNum = findColumns(evolution, colName);
+      } catch (IllegalArgumentException e) {
+        LOG.debug("{}", e.getMessage());
+      }
+      result[i] = colNum;
     }
     return result;
   }
@@ -143,10 +201,8 @@ public class RecordReaderImpl implements RecordReader {
     OrcFile.WriterVersion writerVersion = fileReader.getWriterVersion();
     SchemaEvolution evolution;
     if (options.getSchema() == null) {
-      if (LOG.isInfoEnabled()) {
-        LOG.info("Reader schema not provided -- using file schema " +
-            fileReader.getSchema());
-      }
+      LOG.info("Reader schema not provided -- using file schema " +
+          fileReader.getSchema());
       evolution = new SchemaEvolution(fileReader.getSchema(), null, options);
     } else {
 
@@ -163,14 +219,19 @@ public class RecordReaderImpl implements RecordReader {
             "file schema:   " + fileReader.getSchema());
       }
     }
+
+    this.noSelectedVector = !options.useSelected();
+    LOG.debug("noSelectedVector={}", this.noSelectedVector);
     this.schema = evolution.getReaderSchema();
     this.path = fileReader.path;
     this.rowIndexStride = fileReader.rowIndexStride;
-    boolean ignoreNonUtf8BloomFilter = OrcConf.IGNORE_NON_UTF8_BLOOM_FILTERS.getBoolean(fileReader.conf);
+    boolean ignoreNonUtf8BloomFilter =
+        OrcConf.IGNORE_NON_UTF8_BLOOM_FILTERS.getBoolean(fileReader.conf);
     ReaderEncryption encryption = fileReader.getEncryption();
     this.fileIncluded = evolution.getFileIncluded();
     SearchArgument sarg = options.getSearchArgument();
-    if (sarg != null && rowIndexStride != 0) {
+    boolean[] rowIndexCols = new boolean[evolution.getFileIncluded().length];
+    if (sarg != null && rowIndexStride > 0) {
       sargApp = new SargApplier(sarg,
           rowIndexStride,
           evolution,
@@ -178,6 +239,7 @@ public class RecordReaderImpl implements RecordReader {
           fileReader.useUTCTimestamp,
           fileReader.writerUsedProlepticGregorian(),
           fileReader.options.getConvertToProlepticGregorian());
+      sargApp.setRowIndexCols(rowIndexCols);
     } else {
       sargApp = null;
     }
@@ -228,17 +290,56 @@ public class RecordReaderImpl implements RecordReader {
       skipCorrupt = OrcConf.SKIP_CORRUPT_DATA.getBoolean(fileReader.conf);
     }
 
+    String[] filterCols = null;
+    Consumer<OrcFilterContext> filterCallBack = null;
+    BatchFilter filter = FilterFactory.createBatchFilter(options,
+                                                         evolution.getReaderBaseSchema(),
+                                                         evolution.isSchemaEvolutionCaseAware(),
+                                                         fileReader.getFileVersion(),
+                                                         false);
+    if (filter != null) {
+      // If a filter is determined then use this
+      filterCallBack = filter;
+      filterCols = filter.getColumnNames();
+    }
+
+    // Map columnNames to ColumnIds
+    SortedSet<Integer> filterColIds = new TreeSet<>();
+    if (filterCols != null) {
+      for (String colName : filterCols) {
+        TypeDescription expandCol = findColumnType(evolution, colName);
+        // If the column is not present in the file then this can be ignored from read.
+        if (expandCol == null || expandCol.getId() == -1) {
+          // Add -1 to filter columns so that the NullTreeReader is invoked during the LEADERS phase
+          filterColIds.add(-1);
+          // Determine the common parent and include these
+          expandCol = findMostCommonColumn(evolution, colName);
+        }
+        while (expandCol != null && expandCol.getId() != -1) {
+          // classify the column and the parent branch as LEAD
+          filterColIds.add(expandCol.getId());
+          rowIndexCols[expandCol.getId()] = true;
+          expandCol = expandCol.getParent();
+        }
+      }
+      this.startReadPhase = TypeReader.ReadPhase.LEADERS;
+      LOG.debug("Using startReadPhase: {} with filter columns: {}", startReadPhase, filterColIds);
+    } else {
+      this.startReadPhase = TypeReader.ReadPhase.ALL;
+    }
+
+    this.rowIndexColsToRead = ArrayUtils.contains(rowIndexCols, true) ? rowIndexCols : null;
     TreeReaderFactory.ReaderContext readerContext =
         new TreeReaderFactory.ReaderContext()
           .setSchemaEvolution(evolution)
+          .setFilterCallback(filterColIds, filterCallBack)
           .skipCorrupt(skipCorrupt)
           .fileFormat(fileReader.getFileVersion())
           .useUTCTimestamp(fileReader.useUTCTimestamp)
           .setProlepticGregorian(fileReader.writerUsedProlepticGregorian(),
               fileReader.options.getConvertToProlepticGregorian())
           .setEncryption(encryption);
-    reader = TreeReaderFactory.createTreeReader(evolution.getReaderSchema(),
-        readerContext);
+    reader = TreeReaderFactory.createRootReader(evolution.getReaderSchema(), readerContext);
     skipBloomFilters = hasBadBloomFilters(fileReader.getFileTail().getFooter());
 
     int columns = evolution.getFileSchema().getMaximumId() + 1;
@@ -247,15 +348,17 @@ public class RecordReaderImpl implements RecordReader {
         new OrcProto.BloomFilterIndex[columns]);
 
     planner = new StripePlanner(evolution.getFileSchema(), encryption,
-        dataReader, writerVersion, ignoreNonUtf8BloomFilter,
-        maxDiskRangeChunkLimit);
+                                dataReader, writerVersion, ignoreNonUtf8BloomFilter,
+                                maxDiskRangeChunkLimit, filterColIds);
 
     try {
       advanceToNextRow(reader, 0L, true);
-    } catch (IOException e) {
+    } catch (Exception e) {
       // Try to close since this happens in constructor.
       close();
-      throw e;
+      long stripeId = stripes.size() == 0 ? 0 : stripes.get(0).getStripeId();
+      throw new IOException(String.format("Problem opening stripe %d footer in %s.",
+          stripeId, path), e);
     }
   }
 
@@ -519,7 +622,8 @@ public class RecordReaderImpl implements RecordReader {
                                            TypeDescription type,
                                            boolean writerUsedProlepticGregorian,
                                            boolean useUTCTimestamp) {
-    ColumnStatistics cs = ColumnStatisticsImpl.deserialize(null, statsProto, writerUsedProlepticGregorian, true);
+    ColumnStatistics cs = ColumnStatisticsImpl.deserialize(
+        null, statsProto, writerUsedProlepticGregorian, true);
     ValueRange range = getValueRange(cs, predicate, useUTCTimestamp);
 
     // files written before ORC-135 stores timestamp wrt to local timezone causing issues with PPD.
@@ -615,7 +719,8 @@ public class RecordReaderImpl implements RecordReader {
 
     result = evaluatePredicateMinMax(predicate, predObj, range);
     if (shouldEvaluateBloomFilter(predicate, result, bloomFilter)) {
-      return evaluatePredicateBloomFilter(predicate, predObj, bloomFilter, range.hasNulls, useUTCTimestamp);
+      return evaluatePredicateBloomFilter(
+          predicate, predObj, bloomFilter, range.hasNulls, useUTCTimestamp);
     } else {
       return result;
     }
@@ -627,14 +732,11 @@ public class RecordReaderImpl implements RecordReader {
     // 1) Bloom filter is available
     // 2) Min/Max evaluation yield YES or MAYBE
     // 3) Predicate is EQUALS or IN list
-    if (bloomFilter != null
-        && result != TruthValue.NO_NULL && result != TruthValue.NO
-        && (predicate.getOperator().equals(PredicateLeaf.Operator.EQUALS)
-            || predicate.getOperator().equals(PredicateLeaf.Operator.NULL_SAFE_EQUALS)
-            || predicate.getOperator().equals(PredicateLeaf.Operator.IN))) {
-      return true;
-    }
-    return false;
+    return bloomFilter != null
+           && result != TruthValue.NO_NULL && result != TruthValue.NO
+           && (predicate.getOperator().equals(PredicateLeaf.Operator.EQUALS)
+               || predicate.getOperator().equals(PredicateLeaf.Operator.NULL_SAFE_EQUALS)
+               || predicate.getOperator().equals(PredicateLeaf.Operator.IN));
   }
 
   private static TruthValue evaluatePredicateMinMax(PredicateLeaf predicate,
@@ -746,7 +848,8 @@ public class RecordReaderImpl implements RecordReader {
         for (Object arg : predicate.getLiteralList()) {
           // if atleast one value in IN list exist in bloom filter, qualify the row group/stripe
           Object predObjItem = getBaseObjectForComparison(predicate.getType(), (Comparable) arg);
-          TruthValue result = checkInBloomFilter(bloomFilter, predObjItem, hasNull, useUTCTimestamp);
+          TruthValue result =
+              checkInBloomFilter(bloomFilter, predObjItem, hasNull, useUTCTimestamp);
           if (result == TruthValue.YES_NO_NULL || result == TruthValue.YES_NO) {
             return result;
           }
@@ -764,11 +867,11 @@ public class RecordReaderImpl implements RecordReader {
     TruthValue result = hasNull ? TruthValue.NO_NULL : TruthValue.NO;
 
     if (predObj instanceof Long) {
-      if (bf.testLong(((Long) predObj).longValue())) {
+      if (bf.testLong((Long) predObj)) {
         result = TruthValue.YES_NO_NULL;
       }
     } else if (predObj instanceof Double) {
-      if (bf.testDouble(((Double) predObj).doubleValue())) {
+      if (bf.testDouble((Double) predObj)) {
         result = TruthValue.YES_NO_NULL;
       }
     } else if (predObj instanceof String || predObj instanceof Text ||
@@ -783,7 +886,8 @@ public class RecordReaderImpl implements RecordReader {
           result = TruthValue.YES_NO_NULL;
         }
       } else {
-        if (bf.testLong(SerializationUtils.convertToUtc(TimeZone.getDefault(), ((Timestamp) predObj).getTime()))) {
+        if (bf.testLong(SerializationUtils.convertToUtc(
+            TimeZone.getDefault(), ((Timestamp) predObj).getTime()))) {
           result = TruthValue.YES_NO_NULL;
         }
       }
@@ -804,9 +908,7 @@ public class RecordReaderImpl implements RecordReader {
       result = TruthValue.YES_NO;
     }
 
-    if (LOG.isDebugEnabled()) {
-      LOG.debug("Bloom filter evaluation: " + result.toString());
-    }
+    LOG.debug("Bloom filter evaluation: {}", result);
 
     return result;
   }
@@ -816,7 +918,7 @@ public class RecordReaderImpl implements RecordReader {
    */
   static class SargCastException extends IllegalArgumentException {
 
-    public SargCastException(String string) {
+    SargCastException(String string) {
       super(string);
     }
   }
@@ -852,10 +954,10 @@ public class RecordReaderImpl implements RecordReader {
         break;
       case DECIMAL:
         if (obj instanceof Boolean) {
-          return new HiveDecimalWritable(((Boolean) obj).booleanValue() ?
+          return new HiveDecimalWritable((Boolean) obj ?
               HiveDecimal.ONE : HiveDecimal.ZERO);
         } else if (obj instanceof Integer) {
-          return new HiveDecimalWritable(((Integer) obj).intValue());
+          return new HiveDecimalWritable((Integer) obj);
         } else if (obj instanceof Long) {
           return new HiveDecimalWritable(((Long) obj));
         } else if (obj instanceof Float || obj instanceof Double ||
@@ -882,10 +984,6 @@ public class RecordReaderImpl implements RecordReader {
           return Double.valueOf(obj.toString());
         } else if (obj instanceof Timestamp) {
           return TimestampUtils.getDouble((Timestamp) obj);
-        } else if (obj instanceof HiveDecimal) {
-          return ((HiveDecimal) obj).doubleValue();
-        } else if (obj instanceof BigDecimal) {
-          return ((BigDecimal) obj).doubleValue();
         }
         break;
       case LONG:
@@ -913,7 +1011,7 @@ public class RecordReaderImpl implements RecordReader {
         } else if (obj instanceof Float) {
           return TimestampUtils.doubleToTimestamp(((Float) obj).doubleValue());
         } else if (obj instanceof Double) {
-          return TimestampUtils.doubleToTimestamp(((Double) obj).doubleValue());
+          return TimestampUtils.doubleToTimestamp((Double) obj);
         } else if (obj instanceof HiveDecimal) {
           return TimestampUtils.decimalToTimestamp((HiveDecimal) obj);
         } else if (obj instanceof HiveDecimalWritable) {
@@ -939,8 +1037,8 @@ public class RecordReaderImpl implements RecordReader {
   }
 
   public static class SargApplier {
-    public final static boolean[] READ_ALL_RGS = null;
-    public final static boolean[] READ_NO_RGS = new boolean[0];
+    public static final boolean[] READ_ALL_RGS = null;
+    public static final boolean[] READ_NO_RGS = new boolean[0];
 
     private final OrcFile.WriterVersion writerVersion;
     private final SearchArgument sarg;
@@ -948,8 +1046,7 @@ public class RecordReaderImpl implements RecordReader {
     private final int[] filterColumns;
     private final long rowIndexStride;
     // same as the above array, but indices are set to true
-    private final boolean[] sargColumns;
-    private SchemaEvolution evolution;
+    private final SchemaEvolution evolution;
     private final long[] exceptionCount;
     private final boolean useUTCTimestamp;
     private final boolean writerUsedProlepticGregorian;
@@ -981,19 +1078,21 @@ public class RecordReaderImpl implements RecordReader {
       filterColumns = mapSargColumnsToOrcInternalColIdx(sargLeaves,
                                                         evolution);
       this.rowIndexStride = rowIndexStride;
+      this.evolution = evolution;
+      exceptionCount = new long[sargLeaves.size()];
+      this.useUTCTimestamp = useUTCTimestamp;
+    }
+
+    public void setRowIndexCols(boolean[] rowIndexCols) {
       // included will not be null, row options will fill the array with
       // trues if null
-      sargColumns = new boolean[evolution.getFileIncluded().length];
       for (int i : filterColumns) {
         // filter columns may have -1 as index which could be partition
         // column in SARG.
         if (i > 0) {
-          sargColumns[i] = true;
+          rowIndexCols[i] = true;
         }
       }
-      this.evolution = evolution;
-      exceptionCount = new long[sargLeaves.size()];
-      this.useUTCTimestamp = useUTCTimestamp;
     }
 
     /**
@@ -1053,13 +1152,12 @@ public class RecordReaderImpl implements RecordReader {
                   LOG.info("Skipping ORC PPD - " + e.getMessage() + " on "
                       + predicate);
                 } else {
-                  if (LOG.isWarnEnabled()) {
-                    final String reason = e.getClass().getSimpleName() + " when evaluating predicate." +
-                        " Skipping ORC PPD." +
-                        " Stats: " + stats +
-                        " Predicate: " + predicate;
-                    LOG.warn(reason, e);
-                  }
+                  final String reason = e.getClass().getSimpleName() +
+                      " when evaluating predicate." +
+                      " Skipping ORC PPD." +
+                      " Stats: " + stats +
+                      " Predicate: " + predicate;
+                  LOG.warn(reason, e);
                 }
                 boolean hasNoNull = stats.hasHasNull() && !stats.getHasNull();
                 if (predicate.getOperator().equals(PredicateLeaf.Operator.NULL_SAFE_EQUALS)
@@ -1109,11 +1207,15 @@ public class RecordReaderImpl implements RecordReader {
    * @throws IOException
    */
   protected boolean[] pickRowGroups() throws IOException {
-    // if we don't have a sarg or indexes, we read everything
+    // Read the Row Indicies if required
+    if (rowIndexColsToRead != null) {
+      readCurrentStripeRowIndex();
+    }
+
+    // In the absence of SArg all rows groups should be included
     if (sargApp == null) {
       return null;
     }
-    readRowIndex(currentStripe, fileIncluded, sargApp.sargColumns);
     return sargApp.pickRowGroups(stripes.get(currentStripe),
         indexes.getRowGroupIndex(),
         skipBloomFilters ? null : indexes.getBloomFilterKinds(),
@@ -1146,11 +1248,12 @@ public class RecordReaderImpl implements RecordReader {
 
     // if we haven't skipped the whole stripe, read the data
     if (rowInStripe < rowCountInStripe) {
-      planner.readData(indexes, includedRowGroups, false);
-      reader.startStripe(planner);
+      planner.readData(indexes, includedRowGroups, false, startReadPhase);
+      reader.startStripe(planner, startReadPhase);
+      needsFollowColumnsRead = true;
       // if we skipped the first row group, move the pointers forward
       if (rowInStripe != 0) {
-        seekToRowEntry(reader, (int) (rowInStripe / rowIndexStride));
+        seekToRowEntry(reader, (int) (rowInStripe / rowIndexStride), startReadPhase);
       }
     }
   }
@@ -1162,6 +1265,7 @@ public class RecordReaderImpl implements RecordReader {
     // setup the position in the stripe
     rowCountInStripe = stripe.getNumberOfRows();
     rowInStripe = 0;
+    followRowInStripe = 0;
     rowBaseInStripe = 0;
     for (int i = 0; i < currentStripe; ++i) {
       rowBaseInStripe += stripes.get(i).getNumberOfRows();
@@ -1188,6 +1292,15 @@ public class RecordReaderImpl implements RecordReader {
     }
   }
 
+  /**
+   * Determine the RowGroup based on the supplied row id.
+   * @param rowIdx Row for which the row group is being determined
+   * @return Id of the RowGroup that the row belongs to
+   */
+  private int computeRGIdx(long rowIdx) {
+    return rowIndexStride == 0 ? 0 : (int) (rowIdx / rowIndexStride);
+  }
+
   /**
    * Skip over rows that we aren't selecting, so that the next row is
    * one that we will read.
@@ -1195,15 +1308,14 @@ public class RecordReaderImpl implements RecordReader {
    * @param nextRow the row we want to go to
    * @throws IOException
    */
-  private boolean advanceToNextRow(
-      TreeReaderFactory.TreeReader reader, long nextRow, boolean canAdvanceStripe)
+  private boolean advanceToNextRow(BatchReader reader, long nextRow, boolean canAdvanceStripe)
       throws IOException {
     long nextRowInStripe = nextRow - rowBaseInStripe;
     // check for row skipping
     if (rowIndexStride != 0 &&
         includedRowGroups != null &&
         nextRowInStripe < rowCountInStripe) {
-      int rowGroup = (int) (nextRowInStripe / rowIndexStride);
+      int rowGroup = computeRGIdx(nextRowInStripe);
       if (!includedRowGroups[rowGroup]) {
         while (rowGroup < includedRowGroups.length && !includedRowGroups[rowGroup]) {
           rowGroup += 1;
@@ -1226,10 +1338,10 @@ public class RecordReaderImpl implements RecordReader {
     if (nextRowInStripe != rowInStripe) {
       if (rowIndexStride != 0) {
         int rowGroup = (int) (nextRowInStripe / rowIndexStride);
-        seekToRowEntry(reader, rowGroup);
-        reader.skipRows(nextRowInStripe - rowGroup * rowIndexStride);
+        seekToRowEntry(reader, rowGroup, startReadPhase);
+        reader.skipRows(nextRowInStripe - rowGroup * rowIndexStride, startReadPhase);
       } else {
-        reader.skipRows(nextRowInStripe - rowInStripe);
+        reader.skipRows(nextRowInStripe - rowInStripe, startReadPhase);
       }
       rowInStripe = nextRowInStripe;
     }
@@ -1239,30 +1351,122 @@ public class RecordReaderImpl implements RecordReader {
   @Override
   public boolean nextBatch(VectorizedRowBatch batch) throws IOException {
     try {
-      if (rowInStripe >= rowCountInStripe) {
-        currentStripe += 1;
-        if (currentStripe >= stripes.size()) {
-          batch.size = 0;
-          return false;
+      int batchSize;
+      // do...while is required to handle the case where the filter eliminates all rows in the
+      // batch, we never return an empty batch unless the file is exhausted
+      do {
+        if (rowInStripe >= rowCountInStripe) {
+          currentStripe += 1;
+          if (currentStripe >= stripes.size()) {
+            batch.size = 0;
+            return false;
+          }
+          // Read stripe in Memory
+          readStripe();
+          followRowInStripe = rowInStripe;
         }
-        readStripe();
+
+        batchSize = computeBatchSize(batch.getMaxSize());
+        reader.setVectorColumnCount(batch.getDataColumnCount());
+        reader.nextBatch(batch, batchSize, startReadPhase);
+        if (startReadPhase == TypeReader.ReadPhase.LEADERS && batch.size > 0) {
+          // At least 1 row has been selected and as a result we read the follow columns into the
+          // row batch
+          reader.nextBatch(batch,
+                           batchSize,
+                           prepareFollowReaders(rowInStripe, followRowInStripe));
+          followRowInStripe = rowInStripe + batchSize;
+        }
+        rowInStripe += batchSize;
+        advanceToNextRow(reader, rowInStripe + rowBaseInStripe, true);
+        // batch.size can be modified by filter so only batchSize can tell if we actually read rows
+      } while (batchSize != 0 && batch.size == 0);
+
+      if (noSelectedVector) {
+        // In case selected vector is not supported we leave the size to be read size. In this case
+        // the non filter columns might be read selectively, however the filter after the reader
+        // should eliminate rows that don't match predicate conditions
+        batch.size = batchSize;
+        batch.selectedInUse = false;
       }
 
-      int batchSize = computeBatchSize(batch.getMaxSize());
-
-      rowInStripe += batchSize;
-      reader.setVectorColumnCount(batch.getDataColumnCount());
-      reader.nextBatch(batch, batchSize);
-      batch.selectedInUse = false;
-      batch.size = batchSize;
-      advanceToNextRow(reader, rowInStripe + rowBaseInStripe, true);
-      return batch.size  != 0;
+      return batchSize != 0;
     } catch (IOException e) {
       // Rethrow exception with file name in log message
       throw new IOException("Error reading file: " + path, e);
     }
   }
 
+  /**
+   * This method prepares the non-filter column readers for next batch. This involves the following
+   * 1. Determine position
+   * 2. Perform IO if required
+   * 3. Position the non-filter readers
+   *
+   * This method is repositioning the non-filter columns and as such this method shall never have to
+   * deal with navigating the stripe forward or skipping row groups, all of this should have already
+   * taken place based on the filter columns.
+   * @param toFollowRow The rowIdx identifies the required row position within the stripe for
+   *                    follow read
+   * @param fromFollowRow Indicates the current position of the follow read, exclusive
+   * @return the read phase for reading non-filter columns, this shall be FOLLOWERS_AND_PARENTS in
+   * case of a seek otherwise will be FOLLOWERS
+   */
+  private TypeReader.ReadPhase prepareFollowReaders(long toFollowRow, long fromFollowRow)
+    throws IOException {
+    // 1. Determine the required row group and skip rows needed from the RG start
+    int needRG = computeRGIdx(toFollowRow);
+    // The current row is not yet read so we -1 to compute the previously read row group
+    int readRG = computeRGIdx(fromFollowRow - 1);
+    long skipRows;
+    if (needRG == readRG && toFollowRow >= fromFollowRow) {
+      // In case we are skipping forward within the same row group, we compute skip rows from the
+      // current position
+      skipRows = toFollowRow - fromFollowRow;
+    } else {
+      // In all other cases including seeking backwards, we compute the skip rows from the start of
+      // the required row group
+      skipRows = toFollowRow - (needRG * rowIndexStride);
+    }
+
+    // 2. Plan the row group idx for the non-filter columns if this has not already taken place
+    if (needsFollowColumnsRead) {
+      needsFollowColumnsRead = false;
+      planner.readFollowData(indexes, includedRowGroups, needRG, false);
+      reader.startStripe(planner, TypeReader.ReadPhase.FOLLOWERS);
+    }
+
+    // 3. Position the non-filter readers to the required RG and skipRows
+    TypeReader.ReadPhase result = TypeReader.ReadPhase.FOLLOWERS;
+    if (needRG != readRG || toFollowRow < fromFollowRow) {
+      // When having to change a row group or in case of back navigation, seek both the filter
+      // parents and non-filter. This will re-position the parents present vector. This is needed
+      // to determine the number of non-null values to skip on the non-filter columns.
+      seekToRowEntry(reader, needRG, TypeReader.ReadPhase.FOLLOWERS_AND_PARENTS);
+      // skip rows on both the filter parents and non-filter as both have been positioned in the
+      // previous step
+      reader.skipRows(skipRows, TypeReader.ReadPhase.FOLLOWERS_AND_PARENTS);
+      result = TypeReader.ReadPhase.FOLLOWERS_AND_PARENTS;
+    } else if (skipRows > 0) {
+      // in case we are only skipping within the row group, position the filter parents back to the
+      // position of the follow. This is required to determine the non-null values to skip on the
+      // non-filter columns.
+      seekToRowEntry(reader, readRG, TypeReader.ReadPhase.LEADER_PARENTS);
+      reader.skipRows(fromFollowRow - (readRG * rowIndexStride),
+          TypeReader.ReadPhase.LEADER_PARENTS);
+      // Move both the filter parents and non-filter forward, this will compute the correct
+      // non-null skips on follow children
+      reader.skipRows(skipRows, TypeReader.ReadPhase.FOLLOWERS_AND_PARENTS);
+      result = TypeReader.ReadPhase.FOLLOWERS_AND_PARENTS;
+    }
+    // Identifies the read level that should be performed for the read
+    // FOLLOWERS_WITH_PARENTS indicates repositioning identifying both non-filter and filter parents
+    // FOLLOWERS indicates read only of the non-filter level without the parents, which is used during
+    // contiguous read. During a contiguous read no skips are needed and the non-null information of
+    // the parent is available in the column vector for use during non-filter read
+    return result;
+  }
+
   private int computeBatchSize(long targetBatchSize) {
     final int batchSize;
     // In case of PPD, batch size should be aware of row group boundaries. If only a subset of row
@@ -1270,22 +1474,29 @@ public class RecordReaderImpl implements RecordReader {
     // within strip). Batch size computed out of marker position makes sure that batch size is
     // aware of row group boundary and will not cause overflow when reading rows
     // illustration of this case is here https://issues.apache.org/jira/browse/HIVE-6287
-    if (rowIndexStride != 0 && includedRowGroups != null && rowInStripe < rowCountInStripe) {
+    if (rowIndexStride != 0
+        && (includedRowGroups != null || startReadPhase != TypeReader.ReadPhase.ALL)
+        && rowInStripe < rowCountInStripe) {
       int startRowGroup = (int) (rowInStripe / rowIndexStride);
-      if (!includedRowGroups[startRowGroup]) {
+      if (includedRowGroups != null && !includedRowGroups[startRowGroup]) {
         while (startRowGroup < includedRowGroups.length && !includedRowGroups[startRowGroup]) {
           startRowGroup += 1;
         }
       }
 
       int endRowGroup = startRowGroup;
-      while (endRowGroup < includedRowGroups.length && includedRowGroups[endRowGroup]) {
+      // We force row group boundaries when dealing with filters. We adjust the end row group to
+      // be the next row group even if more than one are possible selections.
+      if (includedRowGroups != null && startReadPhase == TypeReader.ReadPhase.ALL) {
+        while (endRowGroup < includedRowGroups.length
+               && includedRowGroups[endRowGroup]) {
+          endRowGroup += 1;
+        }
+      } else {
         endRowGroup += 1;
       }
 
-      final long markerPosition =
-          (endRowGroup * rowIndexStride) < rowCountInStripe ? (endRowGroup * rowIndexStride)
-              : rowCountInStripe;
+      final long markerPosition = Math.min((endRowGroup * rowIndexStride), rowCountInStripe);
       batchSize = (int) Math.min(targetBatchSize, (markerPosition - rowInStripe));
 
       if (isLogDebugEnabled && batchSize < targetBatchSize) {
@@ -1330,13 +1541,21 @@ public class RecordReaderImpl implements RecordReader {
     throw new IllegalArgumentException("Seek after the end of reader range");
   }
 
-  public OrcIndex readRowIndex(int stripeIndex, boolean[] included,
-                               boolean[] sargColumns) throws IOException {
-    // if this is the current stripe, use the cached objects.
-    if (stripeIndex == currentStripe) {
-      return planner.readRowIndex(
-          sargColumns != null || sargApp == null
-              ? sargColumns : sargApp.sargColumns, indexes);
+  private void readCurrentStripeRowIndex() throws IOException {
+    planner.readRowIndex(rowIndexColsToRead, indexes);
+  }
+
+  public OrcIndex readRowIndex(int stripeIndex,
+                               boolean[] included,
+                               boolean[] readCols) throws IOException {
+    // Use the cached objects if the read request matches the cached request
+    if (stripeIndex == currentStripe
+            && (readCols == null || Arrays.equals(readCols, rowIndexColsToRead))) {
+      if (rowIndexColsToRead != null) {
+        return indexes;
+      } else {
+        return planner.readRowIndex(readCols, indexes);
+      }
     } else {
       StripePlanner copy = new StripePlanner(planner);
       if (included == null) {
@@ -1344,11 +1563,11 @@ public class RecordReaderImpl implements RecordReader {
         Arrays.fill(included, true);
       }
       copy.parseStripe(stripes.get(stripeIndex), included);
-      return copy.readRowIndex(sargColumns, null);
+      return copy.readRowIndex(readCols, null);
     }
   }
 
-  private void seekToRowEntry(TreeReaderFactory.TreeReader reader, int rowEntry)
+  private void seekToRowEntry(BatchReader reader, int rowEntry, TypeReader.ReadPhase readPhase)
       throws IOException {
     OrcProto.RowIndex[] rowIndices = indexes.getRowGroupIndex();
     PositionProvider[] index = new PositionProvider[rowIndices.length];
@@ -1363,7 +1582,7 @@ public class RecordReaderImpl implements RecordReader {
         }
       }
     }
-    reader.seek(index);
+    reader.seek(index, readPhase);
   }
 
   @Override
@@ -1384,7 +1603,10 @@ public class RecordReaderImpl implements RecordReader {
       currentStripe = rightStripe;
       readStripe();
     }
-    readRowIndex(currentStripe, fileIncluded, sargApp == null ? null : sargApp.sargColumns);
+    if (rowIndexColsToRead == null) {
+      // Read the row indexes only if they were not already read as part of readStripe()
+      readCurrentStripeRowIndex();
+    }
 
     // if we aren't to the right row yet, advance in the stripe.
     advanceToNextRow(reader, rowNumber, true);
