@@ -18,9 +18,15 @@
 
 package org.apache.orc.impl;
 
-import static org.junit.Assert.assertEquals;
-import static org.junit.Assert.fail;
+import org.apache.hadoop.hive.common.io.DiskRangeList;
+import org.apache.orc.CompressionCodec;
+import org.apache.orc.EncryptionAlgorithm;
+import org.apache.orc.OrcProto;
+import org.apache.orc.PhysicalWriter;
+import org.apache.orc.impl.writer.StreamOptions;
+import org.junit.jupiter.api.Test;
 
+import javax.crypto.spec.SecretKeySpec;
 import java.io.DataInputStream;
 import java.io.DataOutputStream;
 import java.io.IOException;
@@ -29,14 +35,12 @@ import java.security.Key;
 import java.util.ArrayList;
 import java.util.Arrays;
 import java.util.List;
-import javax.crypto.spec.SecretKeySpec;
-import org.apache.hadoop.hive.common.io.DiskRangeList;
-import org.apache.orc.CompressionCodec;
-import org.apache.orc.EncryptionAlgorithm;
-import org.apache.orc.OrcProto;
-import org.apache.orc.PhysicalWriter;
-import org.apache.orc.impl.writer.StreamOptions;
-import org.junit.Test;
+
+import static org.junit.jupiter.api.Assertions.assertEquals;
+import static org.junit.jupiter.api.Assertions.assertNotEquals;
+import static org.junit.jupiter.api.Assertions.assertNotSame;
+import static org.junit.jupiter.api.Assertions.assertSame;
+import static org.junit.jupiter.api.Assertions.fail;
 
 public class TestInStream {
 
@@ -142,11 +146,11 @@ public class TestInStream {
         in.toString());
     for(int i=0; i < 1024; ++i) {
       int x = in.read();
-      assertEquals("value " + i, i & 0xff, x);
+      assertEquals(i & 0xff, x, "value " + i);
     }
     for(int i=1023; i >= 0; --i) {
       in.seek(positions[i]);
-      assertEquals("value " + i, i & 0xff, in.read());
+      assertEquals(i & 0xff, in.read(), "value " + i);
     }
   }
 
@@ -212,11 +216,11 @@ public class TestInStream {
                        " range: 0 offset: 0 position: 0 limit: 1965",
           in.toString());
       for (int i = 0; i < ROW_COUNT; ++i) {
-        assertEquals("row " + i, i * DATA_CONST, inputStream.readLong());
+        assertEquals(i * DATA_CONST, inputStream.readLong(), "row " + i);
       }
       for (int i = ROW_COUNT - 1; i >= 0; --i) {
         in.seek(positions[i]);
-        assertEquals("row " + i, i * DATA_CONST, inputStream.readLong());
+        assertEquals(i * DATA_CONST, inputStream.readLong(), "row " + i);
       }
     }
   }
@@ -260,11 +264,11 @@ public class TestInStream {
                        " range: 0 offset: 0 position: 0 limit: 1965",
           in.toString());
       for (int i = 0; i < ROW_COUNT; ++i) {
-        assertEquals("row " + i, i * DATA_CONST, inputStream.readLong());
+        assertEquals(i * DATA_CONST, inputStream.readLong(), "row " + i);
       }
       for (int i = ROW_COUNT - 1; i >= 0; --i) {
         in.seek(positions[i]);
-        assertEquals("row " + i, i * DATA_CONST, inputStream.readLong());
+        assertEquals(i * DATA_CONST, inputStream.readLong(), "row " + i);
       }
     }
   }
@@ -339,11 +343,11 @@ public class TestInStream {
                        (bytes.length - 15) + " to " + bytes.length,
           in.toString());
       for (int i = 0; i < ROW_COUNT; ++i) {
-        assertEquals("row " + i, i * DATA_CONST, inputStream.readLong());
+        assertEquals(i * DATA_CONST, inputStream.readLong(), "row " + i);
       }
       for (int i = ROW_COUNT - 1; i >= 0; --i) {
         in.seek(positions[i]);
-        assertEquals("row " + i, i * DATA_CONST, inputStream.readLong());
+        assertEquals(i * DATA_CONST, inputStream.readLong(), "row " + i);
       }
     }
   }
@@ -393,11 +397,11 @@ public class TestInStream {
                        "  range 1 = 2100 to 4044;  range 2 = 4044 to 5044",
           in.toString());
       for (int i = 0; i < ROW_COUNT; ++i) {
-        assertEquals("row " + i, i * DATA_CONST, inputStream.readLong());
+        assertEquals(i * DATA_CONST, inputStream.readLong(), "row " + i);
       }
       for (int i = ROW_COUNT - 1; i >= 0; --i) {
         in.seek(positions[i]);
-        assertEquals("row " + i, i * DATA_CONST, inputStream.readLong());
+        assertEquals(i * DATA_CONST, inputStream.readLong(), "row " + i);
       }
     }
   }
@@ -441,6 +445,114 @@ public class TestInStream {
     }
   }
 
+  private long seekPosition(long prevPos,
+                            PositionCollector[] positions,
+                            int posIdx,
+                            InStream in,
+                            boolean needsSeek)
+    throws IOException {
+    if (needsSeek) {
+      assertNotEquals(prevPos, positions[posIdx].getNext());
+    } else {
+      assertEquals(prevPos, positions[posIdx].getNext());
+    }
+    positions[posIdx].reset();
+    ByteBuffer c = ((InStream.CompressedStream) in).compressed;
+    in.seek(positions[posIdx]);
+    assertEquals(posIdx & 0xff, in.read());
+    if (needsSeek) {
+      assertNotSame(c, ((InStream.CompressedStream) in).compressed);
+    } else {
+      assertSame(c, ((InStream.CompressedStream) in).compressed);
+    }
+    positions[posIdx].reset();
+    return positions[posIdx].getNext();
+  }
+
+  @Test
+  public void testCompressedSeeks() throws Exception {
+    // We test two scenarios one where the stream is perfectly aligned with the DiskRange and the
+    // other where it requires an offset
+    for (int offset : new int[]{0, 10}) {
+      int compValues = 1024;
+      int origValues = 100;
+      PositionCollector[] positions = new PositionCollector[compValues + origValues];
+      byte[] compBytes = getCompressed(positions);
+      assertEquals(961, compBytes.length);
+      // Add an original chunk at the end
+      byte[] bytes = new byte[compBytes.length + 3 + origValues + offset];
+      System.arraycopy(compBytes, 0, bytes, offset, compBytes.length);
+      int startPos = offset + compBytes.length;
+      // Write original header
+      bytes[startPos] = (byte) ((origValues << 1) + 1);
+      bytes[startPos + 1] = (byte) (origValues >> 7);
+      bytes[startPos + 2] = (byte) (origValues >> 15);
+      for (int i = 0; i < 100; i++) {
+        positions[compValues + i] = new PositionCollector();
+        positions[compValues + i].addPosition(compBytes.length);
+        positions[compValues + i].addPosition(i);
+        bytes[startPos + 3 + i] = (byte) (compValues + i);
+      }
+      InStream in = InStream.create("test", new BufferChunk(ByteBuffer.wrap(bytes), 0), offset,
+                                    compBytes.length + 3 + origValues,
+                                    InStream.options()
+                                      .withCodec(new ZlibCodec())
+                                      .withBufferSize(300));
+      assertEquals("compressed stream test position: 0 length: 1064 range: 0" +
+                   String.format(" offset: %d limit: %d range 0 = 0 to %d",
+                                 offset,
+                                 bytes.length,
+                                 bytes.length),
+                   in.toString());
+
+      // Position to the last
+      long currPos = positions[positions.length - 1].getNext();
+      positions[positions.length - 1].reset();
+      in.seek(positions[positions.length - 1]);
+
+      // Seek to the first should reposition compressed
+      currPos = seekPosition(currPos, positions, 0, in, true);
+      // Seek to next position should not require a seek
+      currPos = seekPosition(currPos, positions, 1, in, false);
+
+      // Seek to 301 which should require a seek
+      currPos = seekPosition(currPos, positions, 301, in, true);
+      // Seek to next position should not require a seek
+      seekPosition(currPos, positions, 302, in, false);
+
+      // Seek to 601 which should require a seek
+      currPos = seekPosition(currPos, positions, 601, in, true);
+      // Seek to next position should not require a seek
+      seekPosition(currPos, positions, 602, in, false);
+
+      // Seek to 1024 which should seek to original
+      currPos = seekPosition(currPos, positions, 1024, in, true);
+      // Seek to next position should not require a seek
+      seekPosition(currPos, positions, 1025, in, false);
+      seekPosition(currPos, positions, 1026, in, false);
+    }
+  }
+
+  @Test
+  public void testInvalidSeek() throws Exception {
+    PositionCollector[] positions = new PositionCollector[1024];
+    byte[] bytes = getCompressed(positions);
+
+    assertEquals(961, bytes.length);
+    InStream in = InStream.create("test", new BufferChunk(ByteBuffer.wrap(bytes), 0), 0,
+                                  bytes.length, InStream.options().withCodec(new ZlibCodec()).withBufferSize(300));
+    assertEquals("compressed stream test position: 0 length: 961 range: 0" +
+                 " offset: 0 limit: 961 range 0 = 0 to 961",
+                 in.toString());
+
+    PositionCollector invalidPosition = new PositionCollector();
+    invalidPosition.addPosition(-1);
+    invalidPosition.addPosition(0);
+    in.seek(invalidPosition);
+    assertEquals(0, in.read());
+    assertEquals(1, in.read());
+  }
+
   @Test
   public void testCompressedPartial() throws Exception {
     PositionCollector[] positions = new PositionCollector[1024];
@@ -732,7 +844,7 @@ public class TestInStream {
     byte[] inBuffer = new byte[4096];
     assertEquals(4096, inStream.read(inBuffer));
     for(int i=0; i < inBuffer.length; ++i) {
-      assertEquals("position " + i, (byte)i, inBuffer[i]);
+      assertEquals((byte)i, inBuffer[i], "position " + i);
     }
   }
 
@@ -778,10 +890,9 @@ public class TestInStream {
     posn = 0;
     int read = inStream.read(inBuffer);
     while (read != -1) {
-      assertEquals("Read length at " + posn,
-          Math.min(STREAM_LENGTH - posn, CHUNK_LENGTH), read);
+      assertEquals(Math.min(STREAM_LENGTH - posn, CHUNK_LENGTH), read, "Read length at " + posn);
       for(int i=0; i < read; ++i) {
-        assertEquals("posn " + posn + " + " + i, (byte)(posn + i), inBuffer[i]);
+        assertEquals((byte)(posn + i), inBuffer[i], "posn " + posn + " + " + i);
       }
       posn += read;
       read = inStream.read(inBuffer);
@@ -822,7 +933,7 @@ public class TestInStream {
     byte[] inBuffer = new byte[4096];
     assertEquals(4096, inStream.read(inBuffer));
     for(int i=0; i < inBuffer.length; ++i) {
-      assertEquals("position " + i, (byte)i, inBuffer[i]);
+      assertEquals((byte)i, inBuffer[i], "position " + i);
     }
   }
 
@@ -843,7 +954,7 @@ public class TestInStream {
     byte[] inBuffer = new byte[4096];
     assertEquals(4096, inStream.read(inBuffer));
     for(int i=0; i < inBuffer.length; ++i) {
-      assertEquals("position " + i, (byte)i, inBuffer[i]);
+      assertEquals((byte)i, inBuffer[i], "position " + i);
     }
   }
 }
