@@ -29,14 +29,16 @@ import org.apache.orc.OrcFile;
 import org.apache.orc.Reader;
 import org.apache.orc.TypeDescription;
 import org.apache.orc.Writer;
-import org.junit.After;
-import org.junit.Before;
-import org.junit.Test;
+import org.apache.orc.*;
+import org.junit.jupiter.api.AfterEach;
+import org.junit.jupiter.api.BeforeEach;
+import org.junit.jupiter.api.Disabled;
+import org.junit.jupiter.api.Test;
 
 import java.io.IOException;
-import java.util.Collections;
 
-import static org.junit.Assert.assertEquals;
+import static org.junit.jupiter.api.Assertions.assertEquals;
+import static org.junit.jupiter.api.Assertions.assertThrows;
 
 public class TestWriterImpl {
 
@@ -46,7 +48,7 @@ public class TestWriterImpl {
   Path testFilePath;
   TypeDescription schema;
 
-  @Before
+  @BeforeEach
   public void openFileSystem() throws Exception {
     conf = new Configuration();
     fs = FileSystem.getLocal(conf);
@@ -56,16 +58,18 @@ public class TestWriterImpl {
     schema = TypeDescription.fromString("struct<x:int,y:int>");
   }
 
-  @After
+  @AfterEach
   public void deleteTestFile() throws Exception {
     fs.delete(testFilePath, false);
   }
 
-  @Test(expected = IOException.class)
+  @Test
   public void testDefaultOverwriteFlagForWriter() throws Exception {
-    // default value of the overwrite flag is false, so this should fail
-    Writer w = OrcFile.createWriter(testFilePath, OrcFile.writerOptions(conf).setSchema(schema));
-    w.close();
+    assertThrows(IOException.class, () -> {
+      // default value of the overwrite flag is false, so this should fail
+      Writer w = OrcFile.createWriter(testFilePath, OrcFile.writerOptions(conf).setSchema(schema));
+      w.close();
+    });
   }
 
   @Test
@@ -90,6 +94,33 @@ public class TestWriterImpl {
     w.close();
   }
 
+  @Disabled("ORC-1343: Disable ENABLE_INDEXES tests until reader supports it properly")
+  @Test
+  public void testNoIndexIfEnableIndexIsFalse() throws Exception {
+    conf.set(OrcConf.OVERWRITE_OUTPUT_FILE.getAttribute(), "true");
+    conf.set(OrcConf.ROW_INDEX_STRIDE.getAttribute(), "1000");
+    conf.setBoolean(OrcConf.ENABLE_INDEXES.getAttribute(), false);
+    VectorizedRowBatch b = schema.createRowBatch();
+    LongColumnVector f1 = (LongColumnVector) b.cols[0];
+    LongColumnVector f2 = (LongColumnVector) b.cols[1];
+    Writer w = OrcFile.createWriter(testFilePath, OrcFile.writerOptions(conf).setSchema(schema));
+    long rowCount = 1000;
+    for (int i = 0; i < rowCount; i++) {
+      f1.vector[b.size] = 1 ;
+      f2.vector[b.size] = 2 ;
+      b.size += 1;
+      if (b.size == 10) {
+        w.addRowBatch(b);
+        b.reset();
+      }
+    }
+    w.close();
+
+    for (StripeInformation information: w.getStripes()) {
+      assertEquals(0, information.getIndexLength());
+    }
+  }
+
   @Test
   public void testStripes() throws Exception {
     conf.set(OrcConf.OVERWRITE_OUTPUT_FILE.getAttribute(), "true");
@@ -116,4 +147,12 @@ public class TestWriterImpl {
     Reader r = OrcFile.createReader(testFilePath, OrcFile.readerOptions(conf));
     assertEquals(r.getStripes(), w.getStripes());
   }
+
+  @Test
+  public void testCloseIsIdempotent() throws IOException {
+    conf.set(OrcConf.OVERWRITE_OUTPUT_FILE.getAttribute(), "true");
+    Writer w = OrcFile.createWriter(testFilePath, OrcFile.writerOptions(conf).setSchema(schema));
+    w.close();
+    w.close();
+  }
 }
